/*****************************************************************************
 *
 * FILE:	date.c
 * DESCRIPTION:	time/date utilities for simple CGI library
 * DATE:	Sat, Mar 16 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: date.c,v 1.2 2002/04/06 20:19:58 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include <time.h>
#if	TIME_WITH_SYS_TIME
#include <sys/time.h>
#endif	/* TIME_WITH_SYS_TIME */
#include "cockatrice.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#if	0
#define	JP_LOCALE	1
#endif
typedef struct {
  String	today;
  int		year;
  int		month;	/* 1 .. 12 */
  int		day;	/* 1 .. 31 */
  int		week;	/* 0 .. 6 (0:Sun ... 6:Sat) */
} Date;


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static int	get2ndMonday(int, int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
#if	JP_LOCALE
static String	wday[] = {"", "", "", "", "", "", ""};
#else
static String	wday[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
#endif	/* JP_LOCALE */
static Date	date;


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	dateInit
 * DESCRIPTION:		initialize date
 * ARGUMENTS:		none
 * RETURN VALUE:	date
 */
void
dateInit(void)
{
  time_t	now;
  struct tm *	jst;
  
  time(&now);
  jst = localtime(&now);

  date.year	= jst->tm_year + 1900;
  date.month	= jst->tm_mon + 1;
  date.day	= jst->tm_mday;
  date.week	= jst->tm_wday;

#if	JP_LOCALE
  (void)asprintf(&date.today, "%4dǯ%02d%02d%s",
		 date.year, date.month, date.day, wday[date.week]);
#else
  (void)asprintf(&date.today, "%4d/%02d/%02d(%s)",
		 date.year, date.month, date.day, wday[date.week]);
#endif	/* JP_LOCALE */
}

String getToday(void)		{ return date.today; }
int getYearOfToday(void)	{ return date.year; }
int getMonthOfToday(void)	{ return date.month; }
int getDayOfToday(void)		{ return date.day; }
int getWDayOfToday(void)	{ return date.week; }


/*
 * FUNCTION NAME:	isFestival
 * DESCRIPTION:		Is a day festival?
 * ARGUMENTS:		year, month, day
 * RETURN VALUE:	Festival: TRUE, otherwise: FALSE
 */
bool_t
isFestival(year, month, day)
	int	year;
	int	month;
	int	day;
{
  int	CoA;	/* Coming-of-Age Day (in January) */
  int	VE;	/* Vernal Equinox Day (1996 - ) (in March) */
  int	AE;	/* Autumnal Equinox Day (1996 - ) (in September) */
  int	PC;	/* Physical Culture Day (in October) */

  CoA	= get2ndMonday(year, 1);
  VE	= (int)(20.712662 + 0.242199 * (double)(year - 1996))
	- (int)((year - 1996) / 4);
  AE	= (int)(23.130023 + 0.242199 * (double)(year - 1996))
  	- (int)((year - 1996) / 4);
  PC	= get2ndMonday(year, 10);

  if ((month ==  1 && day ==  1) ||	/* New Year's Day */
      (month ==  1 && day == CoA) ||	/* Coming-of-Age Day */
      (month ==  2 && day == 11) ||	/* National Foundation Day */
      (month ==  3 && day == VE) ||	/* Vernal Equinox Day */
      (month ==  4 && day == 29) ||	/* Green's Day */
      (month ==  5 && day ==  3) ||	/* Constitution Day */
      (month ==  5 && day ==  4) ||	/* National Holiday */
      (month ==  5 && day ==  5) ||	/* Children's Day */
      (month ==  7 && day == 20) ||	/* Sea's Day */
      (month ==  9 && day == 15) ||	/* Old People's Day */
      (month ==  9 && day == AE) ||	/* Autumnal Equinox Day */
      (month == 10 && day == PC) ||	/* Physical Culture Day */
      (month == 11 && day ==  3) ||	/* Culture Day */
      (month == 11 && day == 23) ||	/* Labor Thanksgiving Day */
      (month == 12 && day == 23)) {	/* The Emperor's Birthday */
    return TRUE;
  }

  return FALSE;
}

/*
 * FUNCTION NAME:	get2ndMonday
 * DESCRIPTION:		calculate the 2nd monday for a Happy Monday
 * ARGUMENTS:		year, month
 * RETURN VALUE:	day of Happy Monday
 */
static int
get2ndMonday(year, month)
	int	year;
	int	month;
{
  return 14 - (getDayOfTheWeek(year, month, 1) + 5) % 7;
}

/*
 * FUNCTION NAME:	getDayOfTheWeek
 * DESCRIPTION:		calculate a day of the week
 * ARGUMENTS:		year, month, day
 * RETURN VALUE:	Successfule: 0:Sunday - 6:Saturday, otherwise: -1
 */
int
getDayOfTheWeek(year, month, day)
	int	year;
	int	month;
	int	day;
{
  if (year > 1582) {
    if (month < 3) {
      year--;
      month += 12;
    }
    return (year + (int)(year / 4) - (int)(year / 100) + (int)(year / 400) +
	    (int)((13 * month + 8) / 5) + day) % 7;
  }
  return -1;
}


/*****************************************************************************
 *****************************************************************************/

/*
 * FUNCTION NAME:	getTime
 * DESCRIPTION:		return present time
 * ARGUMENTS:		none
 * RETURN VALUE:	time
 */
String
getTime(void)
{
  String	buf;
  time_t	now;
  struct tm *	jst;
  int		n;
  
  time(&now);
  jst = localtime(&now);

#if	JP_LOCALE
  n = asprintf(&buf, "%02d%02dʬ%02d",
		jst->tm_hour, jst->tm_min, jst->tm_sec);
#else
  n = asprintf(&buf, "%02d:%02d:%02d",
		jst->tm_hour, jst->tm_min, jst->tm_sec);
#endif	/* JP_LOCALE */

  return n == -1 ? NULL : buf;
}

