/*****************************************************************************
 *
 * FILE:	html401.c
 * DESCRIPTION:	HTML4.01 printer for CGI library
 * DATE:	Wed, Apr  3 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: html401.c,v 1.2 2002/04/05 16:12:04 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "catoblepas.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	PrintAttributeID(x)	printGeneralAttribute("id",(x));
#define	PrintAttributeClass(x)	printGeneralAttribute("class",(x));
#define	PrintAttributeLang(x)	printGeneralAttribute("lang",(x));
#define	PrintAttributeTitle(x)	printGeneralAttribute("title",(x));
#define	PrintAttributeStyle(x)	printGeneralAttribute("style",(x));

typedef enum {
  ANCHOR_HREF,
  ANCHOR_NAME
} anchor_t;


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static void	html401Done(void);

static void	htmlBegin(void);
static void	htmlEnd(void);
static void	htmlAttrInit(void);
static void	headBegin(void);
static void	headEnd(void);
static void	headAttrInit(void);
static void	titleOutput(const String);
static void	metaName(const String, const String);
static void	metaHttpEquiv(const String, const String);
static void	linkOutput(void);
static void	linkAttrInit(void);
static void	baseOutput(const String);
static void	styleBegin(const String, media_t);
static void	styleEnd(void);

static void	bodyBegin(void);
static void	bodyEnd(void);
static void	bodyAttrInit(void);

static void	h1Output(const String);
static void	h1AttrInit(void);
static void	h2Output(const String);
static void	h2AttrInit(void);
static void	h3Output(const String);
static void	h3AttrInit(void);
static void	h4Output(const String);
static void	h4AttrInit(void);
static void	h5Output(const String);
static void	h5AttrInit(void);
static void	h6Output(const String);
static void	h6AttrInit(void);
static void	spanOutput(const String);
static void	spanAttrInit(void);
static void	headings(const String, struct _headings *, const String);
static void	headingsAttrInit(struct _headings *);

static void	divBegin(void);
static void	divEnd(void);
static void	divAttrInit(void);
static void	pBegin(void);
static void	pEnd(void);
static void	pAttrInit(void);
static void	blockBegin(const String, struct _block *);
static void	blockAttrInit(struct _block *);

static void	addressBegin(void);
static void	addressEnd(void);
static void	addressAttrInit(void);
static void	preBegin(void);
static void	preEnd(void);
static void	preAttrInit(void);

static void	brOutput(void);
static void	brAttrInit(void);
static void	hrOutput(void);
static void	hrAttrInit(void);
static void	emOutput(const String);
static void	emAttrInit(void);
static void	strongOutput(const String);
static void	strongAttrInit(void);
static void	modifierOutput(const String, struct _modifier *, const String);
static void	modifierAttrInit(struct _modifier *);

static void	imgOutput(const String, const String, Length, Length);
static void	imgAttrInit(void);

static void	ulBegin(void);
static void	ulEnd(void);
static void	ulAttrInit(void);
static void	olBegin(void);
static void	olEnd(void);
static void	olAttrInit(void);
static void	liBegin(void);
static void	liEnd(void);
static void	liAttrInit(void);

static void	tableBegin(void);
static void	tableEnd(void);
static void	tableAttrInit(void);
static void	captionOutput(const String);
static void	captionAttrInit(void);
static void	trBegin(void);
static void	trEnd(void);
static void	trAttrInit(void);
static void	thBegin(void);
static void	thEnd(void);
static void	thAttrInit(void);
static void	tdBegin(void);
static void	tdEnd(void);
static void	tdAttrInit(void);
static void	cellBegin(const String, struct _cell *);
static void	cellAttrInit(struct _cell *);

static void	aHref(const String, const String);
static void	aName(const String, const String);
static void	anchor(anchor_t, const String, const String);
static void	aAttrInit(void);

static void	formBegin(const String, method_t);
static void	formEnd(void);
static void	formAttrInit(void);
static void	multipart_formBegin(const String, method_t);
static void	multipart_formAttrInit(void);
static void	formsBegin(const String, method_t, bool_t, struct _form *);
static void	formsAttrInit(struct _form *);
static void	inputText(const String, const String, Pixels, Pixels);
static void	inputPassword(const String, const String, Pixels, Pixels);
static void	inputCheckbox(const String, const String, bool_t);
static void	inputRadio(const String, const String, bool_t);
static void	inputSubmit(const String, const String);
static void	inputReset(const String, const String);
static void	inputFile(const String, const String, Pixels, Pixels);
static void	inputHidden(const String, const String);
static void	inputImage(const String, const String, const String);
static void	inputButton(const String, const String);
static void	printInputTextGroup(const String, const String, const String,
				    Pixels, Pixels); 
static void	printInputCheckGroup(const String, const String, const String,
				     bool_t);
static void	printInputOtherGroup(const String, const String, const String);
static void	printInputAttr(void);
static void	inputAttrInit(void);
static void	selectBegin(const String, Pixels, bool_t);
static void	selectEnd(void);
static void	selectAttrInit(void);
static void	optionOutput(const String, const String, bool_t);
static void	optionAttrInit(void);
static void	textareaBegin(const String, Number, Number);
static void	textareaEnd(void);
static void	textareaAttrInit(void);

/* Transitional */
static void	fontOutput(const String, const String, const String);
static void	centerBegin(void);
static void	centerEnd(void);
static void	iframeBegin(const String, Length, Length);
static void	iframeEnd(void);
static void	iframeAttrInit(void);

/* Frameset */
static void	framesetBegin(const String, const String);
static void	framesetEnd(void);
static void	frameOutput(const String);
static void	frameAttrInit(void);
static void	noframesBegin(void);
static void	noframesEnd(void);

/**/
static void	Print(const String);

/**/
static void	printGeneralAttribute(const String, const String);
static void	printLinkType(const String, link_t);
static void	printMediaType(media_t);
static void	printTextAlign(align_t);
static void	printHorizontalAlign(align_t);
static void	printImageAlign(align_t);
static void	printCaptionAlign(align_t);
static void	printTableHorizontalAlign(align_t);
static void	printTableVerticalAlign(align_t);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static HTML *	html401;	/* HTML4.01 object */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * function name:	newHTML401
 * description:		create new HTML4.01 object
 * arguments:		DTD type
 * return value:	Successfule: HTML4.01 object, otherwise: NULL
 */
HTML *
newHTML401(dtd)
	dtd_t	dtd;
{
  html401 = (HTML *)calloc(1, sizeof(HTML));
  if (html401 == NULL) {
    return NULL;
  }

  /* DTD */
  html401->dtd = dtd;
  /* HTML */
  html401->html.begin		= (void *)(htmlBegin);
  html401->html.end		= (void *)(htmlEnd);
  html401->html.attr.init	= (void *)(htmlAttrInit);
  /* HEAD */
  html401->head.begin		= (void *)(headBegin);
  html401->head.end		= (void *)(headEnd);
  html401->head.attr.init	= (void *)(headAttrInit);
  /* TITLE */
  html401->title		= (void *)(titleOutput);
  /* META */
  html401->meta.name		= (void *)(metaName);
  html401->meta.http_equiv	= (void *)(metaHttpEquiv);
  /* LINK */
  html401->link.output		= (void *)(linkOutput);
  html401->link.attr.init	= (void *)(linkAttrInit);
  /* BASE */
  html401->base			= (void *)(baseOutput);
  /* STYLE */
  html401->style.begin		= (void *)(styleBegin);
  html401->style.end		= (void *)(styleEnd);

  /* BODY */
  html401->body.begin		= (void *)(bodyBegin);
  html401->body.end		= (void *)(bodyEnd);
  html401->body.attr.init	= (void *)(bodyAttrInit);

  /* H1..H6 */
  html401->h1.output		= (void *)(h1Output);
  html401->h1.attr.init		= (void *)(h1AttrInit);
  html401->h2.output		= (void *)(h2Output);
  html401->h2.attr.init		= (void *)(h2AttrInit);
  html401->h3.output		= (void *)(h3Output);
  html401->h3.attr.init		= (void *)(h3AttrInit);
  html401->h4.output		= (void *)(h4Output);
  html401->h4.attr.init		= (void *)(h4AttrInit);
  html401->h5.output		= (void *)(h5Output);
  html401->h5.attr.init		= (void *)(h5AttrInit);
  html401->h6.output		= (void *)(h6Output);
  html401->h6.attr.init		= (void *)(h6AttrInit);
  /* SPAN */
  html401->span.output		= (void *)(spanOutput);
  html401->span.attr.init	= (void *)(spanAttrInit);

  /* DIV */
  html401->div.begin		= (void *)(divBegin);
  html401->div.end		= (void *)(divEnd);
  html401->div.attr.init	= (void *)(divAttrInit);
  /* P */
  html401->p.begin		= (void *)(pBegin);
  html401->p.end		= (void *)(pEnd);
  html401->p.attr.init		= (void *)(pAttrInit);
  /* ADDRESS */
  html401->address.begin	= (void *)(addressBegin);
  html401->address.end		= (void *)(addressEnd);
  html401->address.attr.init	= (void *)(addressAttrInit);
  /* PRE */
  html401->pre.begin		= (void *)(preBegin);
  html401->pre.end		= (void *)(preEnd);
  html401->pre.attr.init	= (void *)(preAttrInit);

  /* BR */
  html401->br.output		= (void *)(brOutput);
  html401->br.attr.init		= (void *)(brAttrInit);
  /* EM */
  html401->em.output		= (void *)(emOutput);
  html401->em.attr.init		= (void *)(emAttrInit);
  /* STRONG */
  html401->strong.output	= (void *)(strongOutput);
  html401->strong.attr.init	= (void *)(strongAttrInit);
  /* HR */
  html401->hr.output		= (void *)(hrOutput);
  html401->hr.attr.init		= (void *)(hrAttrInit);
  /* IMG */
  html401->img.output		= (void *)(imgOutput);
  html401->img.attr.init	= (void *)(imgAttrInit);

  /* UL */
  html401->ul.begin		= (void *)(ulBegin);
  html401->ul.end		= (void *)(ulEnd);
  html401->ul.attr.init		= (void *)(ulAttrInit);
  /* OL */
  html401->ol.begin		= (void *)(olBegin);
  html401->ol.end		= (void *)(olEnd);
  html401->ol.attr.init		= (void *)(olAttrInit);
  /* LI */
  html401->li.begin		= (void *)(liBegin);
  html401->li.end		= (void *)(liEnd);
  html401->li.attr.init		= (void *)(liAttrInit);

  /* TABLE */
  html401->table.begin		= (void *)(tableBegin);
  html401->table.end		= (void *)(tableEnd);
  html401->table.attr.init	= (void *)(tableAttrInit);
  /* CAPTION */
  html401->caption.output	= (void *)(captionOutput);
  html401->caption.attr.init	= (void *)(captionAttrInit);
  /* TR */
  html401->tr.begin		= (void *)(trBegin);
  html401->tr.end		= (void *)(trEnd);
  html401->tr.attr.init		= (void *)(trAttrInit);
  /* TH */
  html401->th.begin		= (void *)(thBegin);
  html401->th.end		= (void *)(thEnd);
  html401->th.attr.init		= (void *)(thAttrInit);
  /* TD */
  html401->td.begin		= (void *)(tdBegin);
  html401->td.end		= (void *)(tdEnd);
  html401->td.attr.init		= (void *)(tdAttrInit);

  /* A */
  html401->a.href		= (void *)(aHref);
  html401->a.name		= (void *)(aName);
  html401->a.attr.init		= (void *)(aAttrInit);

  /* FORM */
  html401->form.begin		= (void *)(formBegin);
  html401->form.end		= (void *)(formEnd);
  html401->form.attr.init	= (void *)(formAttrInit);
  html401->multipart_form.begin		= (void *)(multipart_formBegin);
  html401->multipart_form.end		= (void *)(formEnd);
  html401->multipart_form.attr.init	= (void *)(multipart_formAttrInit);
  /* INPUT */
  html401->input.text		= (void *)(inputText);
  html401->input.password	= (void *)(inputPassword);
  html401->input.checkbox	= (void *)(inputCheckbox);
  html401->input.radio		= (void *)(inputRadio);
  html401->input.submit		= (void *)(inputSubmit);
  html401->input.reset		= (void *)(inputReset);
  html401->input.file		= (void *)(inputFile);
  html401->input.hidden		= (void *)(inputHidden);
  html401->input.image		= (void *)(inputImage);
  html401->input.button		= (void *)(inputButton);
  html401->input.attr.init	= (void *)(inputAttrInit);
  /* SELECT */
  html401->select.begin		= (void *)(selectBegin);
  html401->select.end		= (void *)(selectEnd);
  html401->select.attr.init	= (void *)(selectAttrInit);
  /* OPTION */
  html401->option.output	= (void *)(optionOutput);
  html401->option.attr.init	= (void *)(optionAttrInit);
  /* TEXTAREA */
  html401->textarea.begin	= (void *)(textareaBegin);
  html401->textarea.end		= (void *)(textareaEnd);
  html401->textarea.attr.init	= (void *)(textareaAttrInit);

  /*
   * Transitional elements
   */
  /* FONT */
  html401->font			= (void *)(fontOutput);
  /* CENTER */
  html401->center.begin		= (void *)(centerBegin);
  html401->center.end		= (void *)(centerEnd);
  /* IFRAME */
  html401->iframe.begin		= (void *)(iframeBegin);
  html401->iframe.end		= (void *)(iframeEnd);
  html401->iframe.attr.init	= (void *)(iframeAttrInit);

  /*
   * Frameset elements
   */
  /* FRAMESET */
  html401->frameset.begin	= (void *)(framesetBegin);
  html401->frameset.end		= (void *)(framesetEnd);
  /* FRAME */
  html401->frame.output		= (void *)(frameOutput);
  html401->frame.attr.init	= (void *)(frameAttrInit);
  /* NOFRAMES */
  html401->noframes.begin	= (void *)(noframesBegin);
  html401->noframes.end		= (void *)(noframesEnd);

  /**/
  html401->print		= (void *)(Print);
  html401->printf		= (int *)(printf);
  html401->done			= (void *)(html401Done);

  return html401;
}

static void
html401Done(void)
{
  free((HTML *)html401);
}


/*****************************************************************************
 *****************************************************************************/

/* HTML */
static void
htmlBegin(void)
{
  switch (html401->dtd) {
    case HTML4:
      html401->dtd = HTML4_Strict;
    case HTML4_Strict:
      fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01//EN\"\n"
	    "\t\"http://www.w3.org/TR/html4/strict.dtd\">\n", stdout);
      break;
    case HTML4_Frameset:
      fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Frameset//EN\"\n"
	    "\t\"http://www.w3.org/TR/html4/frameset.dtd\">\n", stdout);
      break;
    case HTML4_Transitional:
    default:
      fputs("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\"\n"
	    "\t\"http://www.w3.org/TR/html4/loose.dtd\">\n", stdout);
      break;
  }
  fputs("<html", stdout);
  PrintAttributeLang(html401->html.attr.lang);
  fputs(">\n", stdout);
}
static void htmlEnd(void)	{ fputs("</html>\n", stdout); }
static void
htmlAttrInit(void)
{
  html401->html.attr.lang = NULL;
}

/* HEAD */
static void
headBegin(void)
{
  fputs("<head", stdout);
  PrintAttributeLang(html401->head.attr.lang);
  fputs(">\n", stdout);
}
static void headEnd(void)	{ fputs("</head>\n", stdout); }
static void
headAttrInit(void)
{
  html401->head.attr.lang = NULL;
}

/* TITLE */
static void
titleOutput(title)
	const String	title;
{
  fprintf(stdout, "<title>%s</title>\n", title == NULL ? "No title" : title);
}

/* META */
static void
metaName(name, content)
	const String	name;
	const String	content;
{
  if (name != NULL && content != NULL) {
    fprintf(stdout, "<meta name=\"%s\" content=\"%s\">\n", name, content);
  }
}
static void
metaHttpEquiv(name, content)
	const String	name;
	const String	content;
{
  if (name != NULL && content != NULL) {
    fprintf(stdout, "<meta http-equiv=\"%s\" content=\"%s\">\n", name, content);
  }
}

/* LINK */
static void
linkOutput(void)
{
  fputs("<link", stdout);
  if (html401->link.attr.href != NULL) {
    fprintf(stdout, " href=\"%s\"", html401->link.attr.href);
  }
  if (html401->link.attr.type != NULL) {
    fprintf(stdout, " type=\"%s\"", html401->link.attr.type);
  }
  printLinkType("rel", html401->link.attr.rel);
  printLinkType("rev", html401->link.attr.rev);
  printMediaType(html401->link.attr.media);
  if (html401->link.attr.charset != NULL) {
    fprintf(stdout, " charset=\"%s\"", html401->link.attr.charset);
  }
  fputs(">\n", stdout);
}
static void
linkAttrInit(void)
{
  html401->link.attr.href	= NULL;
  html401->link.attr.type	= NULL;
  html401->link.attr.rel	= LINK_NONE;
  html401->link.attr.rev	= LINK_NONE;
  html401->link.attr.media	= MEDIA_NONE;
  html401->link.attr.charset	= NULL;
}

/* BASE */
static void
baseOutput(uri)
	const String	uri;
{
  if (uri != NULL) {
    fprintf(stdout, "<base href=\"%s\">\n", uri);
  }
}

/* STYLE */
static void
styleBegin(mimeType, media)
	const String	mimeType;
	media_t		media;
{
  if (mimeType != NULL) {
    fprintf(stdout, "<style type=\"%s\"", mimeType);
    printMediaType(media);
    fputs(">\n", stdout);
  }
}
static void styleEnd(void)	{ fputs("</style>\n", stdout); }

/* BODY */
static void
bodyBegin(void)
{
  fputs("<body", stdout);
  if (html401->dtd != HTML4_Strict) {
    if (html401->body.attr.trans.background != NULL) {
      fprintf(stdout, " background=\"%s\"",
	      html401->body.attr.trans.background);
    }
    if (html401->body.attr.trans.bgcolor != NULL) {
      fprintf(stdout, " bgcolor=\"%s\"", html401->body.attr.trans.bgcolor);
    }
    if (html401->body.attr.trans.text != NULL) {
      fprintf(stdout, " text=\"%s\"", html401->body.attr.trans.text);
    }
    if (html401->body.attr.trans.link != NULL) {
      fprintf(stdout, " link=\"%s\"", html401->body.attr.trans.link);
    }
    if (html401->body.attr.trans.vlink != NULL) {
      fprintf(stdout, " vlink=\"%s\"", html401->body.attr.trans.vlink);
    }
    if (html401->body.attr.trans.alink != NULL) {
      fprintf(stdout, " alink=\"%s\"", html401->body.attr.trans.alink);
    }
  }
  PrintAttributeID(html401->body.attr.id);
  PrintAttributeClass(html401->body.attr.class);
  PrintAttributeLang(html401->body.attr.lang);
  PrintAttributeTitle(html401->body.attr.title);
  PrintAttributeStyle(html401->body.attr.style);
  fputs(">\n", stdout);
}
static void bodyEnd(void)	{ fputs("</body>\n", stdout); }
static void
bodyAttrInit(void)
{
  html401->body.attr.trans.background	= NULL;
  html401->body.attr.trans.bgcolor	= NULL;
  html401->body.attr.trans.text		= NULL;
  html401->body.attr.trans.link		= NULL;
  html401->body.attr.trans.vlink	= NULL;
  html401->body.attr.trans.alink	= NULL;
  html401->body.attr.id			= NULL;
  html401->body.attr.class		= NULL;
  html401->body.attr.lang		= NULL;
  html401->body.attr.title		= NULL;
  html401->body.attr.style		= NULL;
}

/* H1..H6 */
static void h1Output(const String text) { headings("h1", &html401->h1, text); }
static void h1AttrInit(void)		{ headingsAttrInit(&html401->h1); }
static void h2Output(const String text) { headings("h2", &html401->h2, text); }
static void h2AttrInit(void)		{ headingsAttrInit(&html401->h2); }
static void h3Output(const String text) { headings("h3", &html401->h3, text); }
static void h3AttrInit(void)		{ headingsAttrInit(&html401->h3); }
static void h4Output(const String text) { headings("h4", &html401->h4, text); }
static void h4AttrInit(void)		{ headingsAttrInit(&html401->h4); }
static void h5Output(const String text) { headings("h5", &html401->h5, text); }
static void h5AttrInit(void)		{ headingsAttrInit(&html401->h5); }
static void h6Output(const String text) { headings("h6", &html401->h6, text); }
static void h6AttrInit(void)		{ headingsAttrInit(&html401->h6); }
/* SPAN */
static void spanOutput(const String text)
				  { headings("span", &html401->span, text); }
static void spanAttrInit(void)		{ headingsAttrInit(&html401->span); }

static void
headings(tag, h, text)
	const String		tag;
	struct _headings *	h;
	const String		text;
{
  fprintf(stdout, "<%s", tag);
  if (html401->dtd != HTML4_Strict) {
    printTextAlign(h->attr.trans.align);
  }
  PrintAttributeID(h->attr.id);
  PrintAttributeClass(h->attr.class);
  PrintAttributeLang(h->attr.lang);
  PrintAttributeTitle(h->attr.title);
  PrintAttributeStyle(h->attr.style);
  fprintf(stdout, ">%s</%s>\n", text, tag);
}
static void
headingsAttrInit(h)
	struct _headings *	h;
{
  h->attr.trans.align	= ALIGN_NONE;
  h->attr.id		= NULL;
  h->attr.class		= NULL;
  h->attr.lang		= NULL;
  h->attr.title		= NULL;
  h->attr.style		= NULL;
}

/* DIV */
static void divBegin(void)	{ blockBegin("div", &html401->div); }
static void divEnd(void)	{ fputs("</div>\n", stdout); }
static void divAttrInit(void)	{ blockAttrInit(&html401->div); }
/* P */
static void pBegin(void)	{ blockBegin("p", &html401->p); }
static void pEnd(void)		{ fputs("</p>\n", stdout); }
static void pAttrInit(void)	{ blockAttrInit(&html401->p); }

static void
blockBegin(tag, b)
	const String	tag;
	struct _block *	b;
{
  fprintf(stdout, "<%s", tag);
  if (html401->dtd != HTML4_Strict) {
    printTextAlign(b->attr.trans.align);
  }
  PrintAttributeID(b->attr.id);
  PrintAttributeClass(b->attr.class);
  PrintAttributeLang(b->attr.lang);
  PrintAttributeTitle(b->attr.title);
  PrintAttributeStyle(b->attr.style);
  fputs(">\n", stdout);
}
static void
blockAttrInit(b)
	struct _block *	b;
{
  b->attr.trans.align	= ALIGN_NONE;
  b->attr.id		= NULL;
  b->attr.class		= NULL;
  b->attr.lang		= NULL;
  b->attr.title		= NULL;
  b->attr.style		= NULL;
}

/* ADDRESS */
static void
addressBegin(void)
{
  fputs("<address", stdout);
  PrintAttributeID(html401->address.attr.id);
  PrintAttributeClass(html401->address.attr.class);
  PrintAttributeLang(html401->address.attr.lang);
  fputs(">\n", stdout);
}
static void	addressEnd(void)	{ fputs("</address>\n", stdout); }
static void
addressAttrInit(void)
{
  html401->address.attr.id	= NULL;
  html401->address.attr.class	= NULL;
  html401->address.attr.lang	= NULL;
}

/* PRE */
static void
preBegin(void)
{
  fputs("<pre", stdout);
  PrintAttributeID(html401->pre.attr.id);
  PrintAttributeClass(html401->pre.attr.class);
  PrintAttributeLang(html401->pre.attr.lang);
  PrintAttributeTitle(html401->pre.attr.title);
  PrintAttributeStyle(html401->pre.attr.style);
  fputs(">\n", stdout);
}
static void	preEnd(void)	{ fputs("</pre>\n", stdout); }
static void
preAttrInit(void)
{
  html401->pre.attr.id		= NULL;
  html401->pre.attr.class	= NULL;
  html401->pre.attr.lang	= NULL;
  html401->pre.attr.title	= NULL;
  html401->pre.attr.style	= NULL;
}

/* BR */
static void
brOutput(void)
{
  fputs("<br", stdout);
  if (html401->dtd != HTML4_Strict) {
    switch (html401->br.attr.trans.clear) {
      case CLEAR_LEFT:
	fputs(" clear=\"left\"", stdout);
	break;
      case CLEAR_RIGHT:
	fputs(" clear=\"right\"", stdout);
	break;
      case CLEAR_ALL:
	fputs(" clear=\"all\"", stdout);
	break;
      case CLEAR_NONE:
      default:
	break;
    }
  }
  PrintAttributeID(html401->br.attr.id);
  PrintAttributeClass(html401->br.attr.class);
  PrintAttributeTitle(html401->br.attr.title);
  PrintAttributeStyle(html401->br.attr.style);
  fputs(">\n", stdout);
}
static void
brAttrInit(void)
{
  html401->br.attr.trans.clear	= CLEAR_NONE;
  html401->br.attr.id		= NULL;
  html401->br.attr.class	= NULL;
  html401->br.attr.title	= NULL;
  html401->br.attr.style	= NULL;
}

/* EM */
static void
emOutput(text)
	const String	text;
{
  modifierOutput("em", &html401->em, text);
}
static void emAttrInit(void)	{ modifierAttrInit(&html401->em); }
/* STRONG */
static void
strongOutput(text)
	const String	text;
{
  modifierOutput("strong", &html401->strong, text);
}
static void strongAttrInit(void)	{ modifierAttrInit(&html401->strong); }

static void
modifierOutput(tag, m, text)
	const String		tag;
	struct _modifier *	m;
	const String		text;
{
  fprintf(stdout, "<%s", tag);
  PrintAttributeID(m->attr.id);
  PrintAttributeClass(m->attr.class);
  PrintAttributeLang(m->attr.lang);
  PrintAttributeTitle(m->attr.title);
  PrintAttributeStyle(m->attr.style);
  fprintf(stdout, ">%s</%s>", text, tag);
}
static void
modifierAttrInit(m)
	struct _modifier *	m;
{
  m->attr.id	= NULL;
  m->attr.class	= NULL;
  m->attr.lang	= NULL;
  m->attr.title	= NULL;
  m->attr.style	= NULL;
}

/* HR */
static void
hrOutput(void)
{
  fputs("<hr", stdout);
  if (html401->dtd != HTML4_Strict) {
    printHorizontalAlign(html401->hr.attr.trans.align);
    if (html401->hr.attr.trans.noshade) {
      fputs(" noshade", stdout);
    }
    if (html401->hr.attr.trans.size > 0) {
      fprintf(stdout, " size=\"%d\"", html401->hr.attr.trans.size);
    }
    if (html401->hr.attr.trans.width != NULL) {
      fprintf(stdout, " width=\"%s\"", html401->hr.attr.trans.width);
    }
  }
  PrintAttributeID(html401->hr.attr.id);
  PrintAttributeClass(html401->hr.attr.class);
  PrintAttributeTitle(html401->hr.attr.title);
  PrintAttributeStyle(html401->hr.attr.style);
  fputs(">\n", stdout);
}
static void
hrAttrInit(void)
{
  html401->hr.attr.trans.align	 = ALIGN_NONE;
  html401->hr.attr.trans.noshade = FALSE;
  html401->hr.attr.trans.size	 = 0;
  html401->hr.attr.trans.width	 = NULL;
  html401->hr.attr.id		 = NULL;
  html401->hr.attr.class	 = NULL;
  html401->hr.attr.title	 = NULL;
  html401->hr.attr.style	 = NULL;
}

/* IMG */
static void
imgOutput(src, alt, width, height)
	const String	src;
	const String	alt;
	Length		width;
	Length		height;
{
  fprintf(stdout, "<img src=\"%s\" alt=\"%s\" width=\"%s\" height=\"%s\"",
	  src, alt, width, height);
  if (html401->dtd != HTML4_Strict) {
    printImageAlign(html401->img.attr.trans.align);
    if (html401->img.attr.trans.border != NULL) {
      fprintf(stdout, " border=\"%s\"", html401->img.attr.trans.border);
    }
    if (html401->img.attr.trans.hspace >= 0) {
      fprintf(stdout, " hspace=\"%d\"", html401->img.attr.trans.hspace);
    }
    if (html401->img.attr.trans.vspace >= 0) {
      fprintf(stdout, " vspace=\"%d\"", html401->img.attr.trans.vspace);
    }
  }
  PrintAttributeID(html401->img.attr.id);
  PrintAttributeClass(html401->img.attr.class);
  PrintAttributeLang(html401->img.attr.lang);
  PrintAttributeTitle(html401->img.attr.title);
  PrintAttributeStyle(html401->img.attr.style);
  fputs(">\n", stdout);
}
static void
imgAttrInit(void)
{
  html401->img.attr.trans.align	 = ALIGN_NONE;
  html401->img.attr.trans.border = NULL;
  html401->img.attr.trans.hspace = -1;
  html401->img.attr.trans.vspace = -1;
  html401->img.attr.id		 = NULL;
  html401->img.attr.class	 = NULL;
  html401->img.attr.lang	 = NULL;
  html401->img.attr.title	 = NULL;
  html401->img.attr.style	 = NULL;
}

/* UL */
static void
ulBegin(void)
{
  fputs("<ul", stdout);
  if (html401->dtd != HTML4_Strict) {
    if (html401->ul.attr.trans.bullet != BULLET_NONE) {
      fputs(" type=", stdout);
      switch (html401->ul.attr.trans.bullet) {
	case BULLET_DISC:
	  fputs("\"disc\"", stdout);
	  break;
	case BULLET_CIRCLE:
	  fputs("\"circle\"", stdout);
	  break;
	case BULLET_SQUARE:
	  fputs("\"square\"", stdout);
	  break;
	default:
	  break;
      }
    }
    if (html401->ul.attr.trans.compact) {
      fputs(" compact", stdout);
    }
  }
  PrintAttributeID(html401->ul.attr.id);
  PrintAttributeClass(html401->ul.attr.class);
  PrintAttributeLang(html401->ul.attr.lang);
  PrintAttributeTitle(html401->ul.attr.title);
  PrintAttributeStyle(html401->ul.attr.style);
  fputs(">\n", stdout);
}
static void ulEnd(void)	{ fputs("</ul>\n", stdout); }
static void
ulAttrInit(void)
{
  html401->ul.attr.trans.bullet	 = BULLET_NONE;
  html401->ul.attr.trans.compact = FALSE;
  html401->ul.attr.id		 = NULL;
  html401->ul.attr.class	 = NULL;
  html401->ul.attr.lang		 = NULL;
  html401->ul.attr.title	 = NULL;
  html401->ul.attr.style	 = NULL;
}

/* OL */
static void
olBegin(void)
{
  fputs("<ol", stdout);
  if (html401->dtd != HTML4_Strict) {
    if (html401->ol.attr.trans.numbering != NUMBERING_NONE) {
      fputs(" type=", stdout);
      switch (html401->ol.attr.trans.numbering) {
	case NUMBERING_ARABIC_NUMBER:
	  fputs("\"1\"", stdout);
	  break;
	case NUMBERING_LOWER_ALPHA:
	  fputs("\"a\"", stdout);
	  break;
	case NUMBERING_UPPER_ALPHA:
	  fputs("\"A\"", stdout);
	  break;
	case NUMBERING_LOWER_ROMAN:
	  fputs("\"i\"", stdout);
	  break;
	case NUMBERING_UPPER_ROMAN:
	  fputs("\"I\"", stdout);
	  break;
	default:
	  break;
      }
    }
    if (html401->ol.attr.trans.start != 1) {
      fprintf(stdout, " start=\"%d\"", html401->ol.attr.trans.start);
    }
    if (html401->ol.attr.trans.compact) {
      fputs(" compact", stdout);
    }
  }
  PrintAttributeID(html401->ol.attr.id);
  PrintAttributeClass(html401->ol.attr.class);
  PrintAttributeLang(html401->ol.attr.lang);
  PrintAttributeTitle(html401->ol.attr.title);
  PrintAttributeStyle(html401->ol.attr.style);
  fputs(">\n", stdout);
}
static void olEnd(void)	{ fputs("</ol>\n", stdout); }
static void
olAttrInit(void)
{
  html401->ol.attr.trans.numbering = NUMBERING_NONE;
  html401->ol.attr.trans.start	 = 1;
  html401->ol.attr.trans.compact = FALSE;
  html401->ol.attr.id		 = NULL;
  html401->ol.attr.class	 = NULL;
  html401->ol.attr.lang		 = NULL;
  html401->ol.attr.title	 = NULL;
  html401->ol.attr.style	 = NULL;
}

/* LI */
static void
liBegin(void)
{
  fputs("<li", stdout);
  if (html401->dtd != HTML4_Strict) {
    if (html401->li.attr.trans.style != STYLE_NONE) {
      fputs(" type=", stdout);
      switch (html401->li.attr.trans.style) {
	case STYLE_DISC:
	  fputs("\"disc\"", stdout);
	  break;
	case STYLE_CIRCLE:
	  fputs("\"circle\"", stdout);
	  break;
	case STYLE_SQUARE:
	  fputs("\"square\"", stdout);
	  break;
	case STYLE_ARABIC_NUMBER:
	  fputs("\"1\"", stdout);
	  break;
	case STYLE_LOWER_ALPHA:
	  fputs("\"a\"", stdout);
	  break;
	case STYLE_UPPER_ALPHA:
	  fputs("\"A\"", stdout);
	  break;
	case STYLE_LOWER_ROMAN:
	  fputs("\"i\"", stdout);
	  break;
	case STYLE_UPPER_ROMAN:
	  fputs("\"I\"", stdout);
	  break;
	default:
	  break;
      }
    }
    if (html401->li.attr.trans.value > 0) {
      fprintf(stdout, " value=\"%d\"", html401->li.attr.trans.value);
    }
  }
  PrintAttributeID(html401->li.attr.id);
  PrintAttributeClass(html401->li.attr.class);
  PrintAttributeLang(html401->li.attr.lang);
  PrintAttributeTitle(html401->li.attr.title);
  PrintAttributeStyle(html401->li.attr.style);
  fputs(">\n", stdout);
}
static void liEnd(void)	{ fputs("</li>\n", stdout); }
static void
liAttrInit(void)
{
  html401->li.attr.trans.style	= STYLE_NONE;
  html401->li.attr.trans.value	= 0;
  html401->li.attr.id		= NULL;
  html401->li.attr.class	= NULL;
  html401->li.attr.lang		= NULL;
  html401->li.attr.title	= NULL;
  html401->li.attr.style	= NULL;
}

/* TABLE */
static void
tableBegin(void)
{
  fputs("<table", stdout);
  if (html401->table.attr.summary != NULL) {
    fprintf(stdout, " summary=\"%s\"", html401->table.attr.summary);
  }
  if (html401->table.attr.width != NULL) {
    fprintf(stdout, " width=\"%s\"", html401->table.attr.width);
  }
  if (html401->table.attr.frame != FRAME_VOID) {
    fputs(" frame=", stdout);
    switch (html401->table.attr.frame) {
      case FRAME_ABOVE:
	fputs("\"above\"", stdout);
	break;
      case FRAME_BELOW:
	fputs("\"below\"", stdout);
	break;
      case FRAME_HSIDES:
	fputs("\"hsides\"", stdout);
	break;
      case FRAME_LHS:
	fputs("\"lhs\"", stdout);
	break;
      case FRAME_RHS:
	fputs("\"rhs\"", stdout);
	break;
      case FRAME_VSIDES:
	fputs("\"vsides\"", stdout);
	break;
      case FRAME_BOX:
	fputs("\"box\"", stdout);
	break;
      case FRAME_BORDER:
	fputs("\"border\"", stdout);
	break;
      default:
	break;
    }
  }
  if (html401->table.attr.rules != RULES_NONE) {
    fputs(" rules=", stdout);
    switch (html401->table.attr.rules) {
      case RULES_GROUPS:
	fputs("\"groups\"", stdout);
	break;
      case RULES_ROWS:
	fputs("\"rows\"", stdout);
	break;
      case RULES_COLS:
	fputs("\"cols\"", stdout);
	break;
      case RULES_ALL:
	fputs("\"all\"", stdout);
	break;
      default:
	break;
    }
  }
  if (html401->table.attr.border >= 0) {
    fprintf(stdout, " border=\"%d\"", html401->table.attr.border);
  }
  if (html401->table.attr.cellspacing != NULL) {
    fprintf(stdout, " cellspacing=\"%s\"", html401->table.attr.cellspacing);
  }
  if (html401->table.attr.cellpadding != NULL) {
    fprintf(stdout, " cellpadding=\"%s\"", html401->table.attr.cellpadding);
  }
  if (html401->dtd != HTML4_Strict) {
    printHorizontalAlign(html401->table.attr.trans.align);
    if (html401->table.attr.trans.bgcolor != NULL) {
      fprintf(stdout, " bgcolor=\"%s\"", html401->table.attr.trans.bgcolor);
    }
  }
  PrintAttributeID(html401->table.attr.id);
  PrintAttributeClass(html401->table.attr.class);
  PrintAttributeLang(html401->table.attr.lang);
  PrintAttributeTitle(html401->table.attr.title);
  PrintAttributeStyle(html401->table.attr.style);
  fputs(">\n", stdout);
}
static void tableEnd(void)	{ fputs("</table>\n", stdout); }
static void
tableAttrInit(void)
{
  html401->table.attr.summary		= NULL;
  html401->table.attr.width		= NULL;
  html401->table.attr.frame		= FRAME_VOID;
  html401->table.attr.rules		= RULES_NONE;
  html401->table.attr.border		= 1;
  html401->table.attr.cellspacing	= NULL;
  html401->table.attr.cellpadding	= NULL;
  html401->table.attr.trans.align	= ALIGN_NONE;
  html401->table.attr.trans.bgcolor	= NULL;
  html401->table.attr.id		= NULL;
  html401->table.attr.class		= NULL;
  html401->table.attr.lang		= NULL;
  html401->table.attr.title		= NULL;
  html401->table.attr.style		= NULL;
}

/* CAPTION */
static void
captionOutput(text)
	const String	text;
{
  fputs("<caption", stdout);
  if (html401->dtd != HTML4_Strict) {
    printCaptionAlign(html401->caption.attr.trans.align);
  }
  PrintAttributeID(html401->caption.attr.id);
  PrintAttributeClass(html401->caption.attr.class);
  PrintAttributeLang(html401->caption.attr.lang);
  PrintAttributeTitle(html401->caption.attr.title);
  PrintAttributeStyle(html401->caption.attr.style);
  fprintf(stdout, ">%s</caption>\n", text);
}
static void
captionAttrInit(void)
{
  html401->caption.attr.trans.align	= ALIGN_NONE;
  html401->caption.attr.id		= NULL;
  html401->caption.attr.class		= NULL;
  html401->caption.attr.lang		= NULL;
  html401->caption.attr.title		= NULL;
  html401->caption.attr.style		= NULL;
}

/* TR */
static void
trBegin(void)
{
  fputs("<tr", stdout);
  printTableHorizontalAlign(html401->tr.attr.align);
  printTableVerticalAlign(html401->tr.attr.valign);
  if (html401->tr.attr.Char != NULL) {
    fprintf(stdout, " char=\"%s\"", html401->tr.attr.Char);
  }
  if (html401->tr.attr.charoff != NULL) {
    fprintf(stdout, " charoff=\"%s\"", html401->tr.attr.charoff);
  }
  if (html401->dtd != HTML4_Strict) {
    if (html401->tr.attr.trans.bgcolor != NULL) {
      fprintf(stdout, " bgcolor=\"%s\"", html401->tr.attr.trans.bgcolor);
    }
  }
  PrintAttributeID(html401->tr.attr.id);
  PrintAttributeClass(html401->tr.attr.class);
  PrintAttributeLang(html401->tr.attr.lang);
  PrintAttributeTitle(html401->tr.attr.title);
  PrintAttributeStyle(html401->tr.attr.style);
  fputs(">\n", stdout);
}
static void trEnd(void)	{ fputs("</tr>\n", stdout); }
static void
trAttrInit(void)
{
  html401->tr.attr.align		= ALIGN_NONE;
  html401->tr.attr.valign		= ALIGN_NONE;
  html401->tr.attr.Char			= NULL;
  html401->tr.attr.charoff		= NULL;
  html401->tr.attr.trans.bgcolor	= NULL;
  html401->tr.attr.id			= NULL;
  html401->tr.attr.class		= NULL;
  html401->tr.attr.lang			= NULL;
  html401->tr.attr.title		= NULL;
  html401->tr.attr.style		= NULL;
}

/* TH */
static void thBegin(void)	{ cellBegin("th", &html401->th); }
static void thEnd(void)		{ fputs("</th>", stdout); }
static void thAttrInit(void)	{ cellAttrInit(&html401->th); }
/* TD */
static void tdBegin(void)	{ cellBegin("td", &html401->td); }
static void tdEnd(void)		{ fputs("</td>", stdout); }
static void tdAttrInit(void)	{ cellAttrInit(&html401->td); }

static void
cellBegin(tag, c)
	const String	tag;
	struct _cell *	c;
{
  fprintf(stdout, "<%s", tag);
  if (c->attr.scope != SCOPE_NONE) {
    fputs(" scope=", stdout);
    switch (c->attr.scope) {
      case SCOPE_ROW:
	fputs("\"row\"", stdout);
	break;
      case SCOPE_COL:
	fputs("\"col\"", stdout);
	break;
      case SCOPE_ROWGROUP:
	fputs("\"rowgroup\"", stdout);
	break;
      case SCOPE_COLGROUP:
	fputs("\"colgroup\"", stdout);
	break;
      default:
	break;
    }
  }
  if (c->attr.abbr != NULL) {
    fprintf(stdout, " abbr=\"%s\"", c->attr.abbr);
  }
  if (c->attr.axis != NULL) {
    fprintf(stdout, " axis=\"%s\"", c->attr.axis);
  }
  if (c->attr.rowspan > 0) {
    fprintf(stdout, " rowspan=\"%d\"", c->attr.rowspan);
  }
  if (c->attr.colspan > 0) {
    fprintf(stdout, " colspan=\"%d\"", c->attr.colspan);
  }
  printTableHorizontalAlign(c->attr.align);
  printTableVerticalAlign(c->attr.valign);
  if (c->attr.Char != NULL) {
    fprintf(stdout, " char=\"%s\"", c->attr.Char);
  }
  if (c->attr.charoff != NULL) {
    fprintf(stdout, " charoff=\"%s\"", c->attr.charoff);
  }
  if (html401->dtd != HTML4_Strict) {
    if (c->attr.trans.nowrap) {
      fputs(" nowrap", stdout);
    }
    if (c->attr.trans.width > 0) {
      fprintf(stdout, " width=\"%d\"", c->attr.trans.width);
    }
    if (c->attr.trans.height > 0) {
      fprintf(stdout, " height=\"%d\"", c->attr.trans.height);
    }
    if (c->attr.trans.bgcolor != NULL) {
      fprintf(stdout, " bgcolor=\"%s\"", c->attr.trans.bgcolor);
    }
  }
  PrintAttributeID(c->attr.id);
  PrintAttributeClass(c->attr.class);
  PrintAttributeLang(c->attr.lang);
  PrintAttributeTitle(c->attr.title);
  PrintAttributeStyle(c->attr.style);
  fputs(">\n", stdout);
}
static void
cellAttrInit(c)
	struct _cell *	c;
{
  c->attr.scope		= SCOPE_NONE;
  c->attr.abbr		= NULL;
  c->attr.axis		= NULL;
  c->attr.rowspan	= 0;
  c->attr.colspan	= 0;
  c->attr.align		= ALIGN_NONE;
  c->attr.valign	= ALIGN_NONE;
  c->attr.Char		= NULL;
  c->attr.charoff	= NULL;
  c->attr.trans.nowrap	= FALSE;
  c->attr.trans.width	= 0;
  c->attr.trans.height	= 0;
  c->attr.trans.bgcolor	= NULL;
  c->attr.id		= NULL;
  c->attr.class		= NULL;
  c->attr.lang		= NULL;
  c->attr.title		= NULL;
  c->attr.style		= NULL;
}

/* A */
static void
aHref(uri, text)
	const String	uri;
	const String	text;
{
  anchor(ANCHOR_HREF, uri, text);
}
static void
aName(name, text)
	const String	name;
	const String	text;
{
  anchor(ANCHOR_NAME, name, text);
}
static void
anchor(anchorType, name, text)
	anchor_t	anchorType;
	const String	name;
	const String	text;
{
  fputs("<a ", stdout);
  switch (anchorType) {
    case ANCHOR_HREF:
      fprintf(stdout, "href=\"%s\"", name);
      break;
    case ANCHOR_NAME:
      fprintf(stdout, "name=\"%s\"", name);
      break;
    default:
      break;
  }
  if (html401->a.attr.type != NULL) {
    fprintf(stdout, " type=\"%s\"", html401->a.attr.type);
  }
  printLinkType("rel", html401->a.attr.rel);
  printLinkType("rev", html401->a.attr.rev);
  if (html401->a.attr.target != NULL) {
    fprintf(stdout, " target=\"%s\"", html401->a.attr.target);
  }
  PrintAttributeID(html401->a.attr.id);
  PrintAttributeClass(html401->a.attr.class);
  PrintAttributeLang(html401->a.attr.lang);
  PrintAttributeTitle(html401->a.attr.title);
  PrintAttributeStyle(html401->a.attr.style);
  fprintf(stdout, ">%s</a>\n", text);
}
static void
aAttrInit(void)
{
  html401->a.attr.type		= NULL;
  html401->a.attr.rel		= LINK_NONE;
  html401->a.attr.rev		= LINK_NONE;
  html401->a.attr.target	= NULL;
  html401->a.attr.id		= NULL;
  html401->a.attr.class		= NULL;
  html401->a.attr.lang		= NULL;
  html401->a.attr.title		= NULL;
  html401->a.attr.style		= NULL;
}

/* FORM */
static void
formBegin(action, method)
	const String	action;
	method_t	method;
{
  if (action != NULL) {
    formsBegin(action, method, FALSE, &html401->form);
  }
}
static void formEnd(void)	{ fputs("</form>\n", stdout); }
static void formAttrInit(void)	{ formsAttrInit(&html401->form); }
static void
multipart_formBegin(action, method)
	const String	action;
	method_t	method;
{
  if (action != NULL) {
    formsBegin(action, method, TRUE, &html401->multipart_form);
  }
}
static void
multipart_formAttrInit(void)
{
  formsAttrInit(&html401->multipart_form);
}
static void
formsBegin(action, method, multipart, f)
	const String	action;
	method_t	method;
	bool_t		multipart;
	struct _form *	f;
{
  fprintf(stdout, "<form action=\"%s\" method=", action);
  switch (method) {
    case METHOD_POST:
      fputs("\"post\" ", stdout);
      break;
    case METHOD_GET:
    default:
      fputs("\"get\" ", stdout);
      break;
  }
  fputs("enctype=", stdout);
  if (multipart) {
    fputs("\"multipart/form-data\" ", stdout);
  }
  else {
    fputs("\"application/x-www-form-urlencoded\" ", stdout);
  }
  PrintAttributeID(f->attr.id);
  PrintAttributeClass(f->attr.class);
  PrintAttributeLang(f->attr.lang);
  PrintAttributeTitle(f->attr.title);
  PrintAttributeStyle(f->attr.style);
  fputs(">\n", stdout);
}
static void
formsAttrInit(f)
	struct _form *	f;
{
  f->attr.id	= NULL;
  f->attr.class	= NULL;
  f->attr.lang	= NULL;
  f->attr.title	= NULL;
  f->attr.style	= NULL;
}

/* INPUT */
static void
inputText(name, value, size, maxlength)
	const String	name;
	const String	value;
	Pixels		size;
	Pixels		maxlength;
{
  printInputTextGroup("text", name, value, size, maxlength);
}
static void
inputPassword(name, value, size, maxlength)
	const String	name;
	const String	value;
	Pixels		size;
	Pixels		maxlength;
{
  printInputTextGroup("password", name, value, size, maxlength);
}
static void
inputCheckbox(name, value, checked)
	const String	name;
	const String	value;
	bool_t		checked;
{
  printInputCheckGroup("checkbox", name, value, checked);
}
static void
inputRadio(name, value, checked)
	const String	name;
	const String	value;
	bool_t		checked;
{
  printInputCheckGroup("radio", name, value, checked);
}
static void
inputSubmit(name, value)
	const String	name;
	const String	value;
{
  printInputOtherGroup("submit", name, value);
}
static void
inputReset(name, value)
	const String	name;
	const String	value;
{
  printInputOtherGroup("reset", name, value);
}
static void
inputFile(name, value, size, maxlength)
	const String	name;
	const String	value;
	Pixels		size;
	Pixels		maxlength;
{
  printInputTextGroup("file", name, value, size, maxlength);
}
static void
inputHidden(name, value)
	const String	name;
	const String	value;
{
  printInputOtherGroup("hidden", name, value);
}
static void
inputImage(name, src, alt)
	const String	name;
	const String	src;
	const String	alt;
{
  fprintf(stdout, "<input type=\"image\" name=\"%s\" src=\"%s\" alt=\"%s\"",
	  name, src, alt);
  if (html401->dtd != HTML4_Strict) {
    printImageAlign(html401->input.attr.trans.align);
  }
  printInputAttr();
  fputs(">\n", stdout);
}
static void
inputButton(name, value)
	const String	name;
	const String	value;
{
  printInputOtherGroup("button", name, value);
}
static void
printInputTextGroup(type, name, value, size, maxlength)
	const String	type;
	const String	name;
	const String	value;
	Pixels		size;
	Pixels		maxlength;
{
  fprintf(stdout, "<input type=\"%s\" name=\"%s\" value=\"%s\"",
	  type, name, value);
  if (size > 0) {
    fprintf(stdout, " size=\"%d\"", size);
  }
  if (maxlength > 0) {
    fprintf(stdout, " maxlength=\"%d\"", maxlength);
  }
  printInputAttr();
  fputs(">\n", stdout);
}
static void
printInputCheckGroup(type, name, value, checked)
	const String	type;
	const String	name;
	const String	value;
	bool_t		checked;
{
  fprintf(stdout, "<input type=\"%s\" name=\"%s\" value=\"%s\"",
	  type, name, value);
  if (checked) {
    fputs(" checked", stdout);
  }
  printInputAttr();
  fputs(">\n", stdout);
}
static void
printInputOtherGroup(type, name, value)
	const String	type;
	const String	name;
	const String	value;
{
  fprintf(stdout, "<input type=\"%s\" name=\"%s\" value=\"%s\"",
	  type, name, value);
  printInputAttr();
  fputs(">\n", stdout);
}
static void
printInputAttr(void)
{
  if (html401->input.attr.readonly) {
    fputs(" readonly", stdout);
  }
  if (html401->input.attr.disabled) {
    fputs(" disabled", stdout);
  }
  PrintAttributeID(html401->input.attr.id);
  PrintAttributeClass(html401->input.attr.class);
  PrintAttributeLang(html401->input.attr.lang);
  PrintAttributeTitle(html401->input.attr.title);
  PrintAttributeStyle(html401->input.attr.style);
}
static void
inputAttrInit(void)
{
  html401->input.attr.readonly		= FALSE;
  html401->input.attr.disabled		= FALSE;
  html401->input.attr.trans.align	= ALIGN_NONE;
  html401->input.attr.id		= NULL;
  html401->input.attr.class		= NULL;
  html401->input.attr.lang		= NULL;
  html401->input.attr.title		= NULL;
  html401->input.attr.style		= NULL;
}

/* SELECT */
static void
selectBegin(name, size, multiple)
	const String	name;
	Pixels		size;
	bool_t		multiple;
{
  fprintf(stdout, "<select name=\"%s\"", name);
  if (size > 0) {
    fprintf(stdout, " size=\"%d\"", size);
  }
  if (multiple) {
    fputs(" multiple", stdout);
  }
  PrintAttributeID(html401->select.attr.id);
  PrintAttributeClass(html401->select.attr.class);
  PrintAttributeLang(html401->select.attr.lang);
  PrintAttributeTitle(html401->select.attr.title);
  PrintAttributeStyle(html401->select.attr.style);
  fputs(">\n", stdout);
}
static void selectEnd(void) { fputs("</select>\n", stdout); }
static void
selectAttrInit(void)
{
  html401->select.attr.id		= NULL;
  html401->select.attr.class		= NULL;
  html401->select.attr.lang		= NULL;
  html401->select.attr.title		= NULL;
  html401->select.attr.style		= NULL;
}

/* option */
static void
optionOutput(text, value, selected)
	const String	text;
	const String	value;
	bool_t		selected;
{
  fputs("<option", stdout);
  if (value != NULL) {
    fprintf(stdout, " value=\"%s\"", value);
  }
  if (selected) {
    fputs(" selected", stdout);
  }
  if (html401->option.attr.label != NULL) {
    fprintf(stdout, " label=\"%s\"", html401->option.attr.label);
  }
  if (selected != FALSE && html401->option.attr.disabled) {
    fputs(" disabled", stdout);
  }
  PrintAttributeID(html401->option.attr.id);
  PrintAttributeClass(html401->option.attr.class);
  PrintAttributeLang(html401->option.attr.lang);
  PrintAttributeTitle(html401->option.attr.title);
  PrintAttributeStyle(html401->option.attr.style);
  fprintf(stdout, ">%s</option>\n", text);
}
static void
optionAttrInit(void)
{
  html401->option.attr.label		= NULL;
  html401->option.attr.disabled		= FALSE;
  html401->option.attr.id		= NULL;
  html401->option.attr.class		= NULL;
  html401->option.attr.lang		= NULL;
  html401->option.attr.title		= NULL;
  html401->option.attr.style		= NULL;
}

/* TEXTAREA */
static void
textareaBegin(name, rows, cols)
	const String	name;
	Number		rows;
	Number		cols;
{
  fputs("<textarea", stdout);
  if (name != NULL) {
    fprintf(stdout, " name=\"%s\"", name);
  }
  if (rows > 0) {
    fprintf(stdout, " rows=\"%d\"", rows);
  }
  if (cols > 0) {
    fprintf(stdout, " cols=\"%d\"", cols);
  }
  if (html401->textarea.attr.readonly) {
    fputs(" readonly", stdout);
  }
  if (html401->textarea.attr.disabled) {
    fputs(" disabled", stdout);
  }
  PrintAttributeID(html401->textarea.attr.id);
  PrintAttributeClass(html401->textarea.attr.class);
  PrintAttributeLang(html401->textarea.attr.lang);
  PrintAttributeTitle(html401->textarea.attr.title);
  PrintAttributeStyle(html401->textarea.attr.style);
  fputs(">", stdout);
}
static void textareaEnd(void)	{ fputs("</textarea>\n", stdout); }
static void
textareaAttrInit(void)
{
  html401->textarea.attr.readonly	= FALSE;
  html401->textarea.attr.disabled	= FALSE;
  html401->textarea.attr.id		= NULL;
  html401->textarea.attr.class		= NULL;
  html401->textarea.attr.lang		= NULL;
  html401->textarea.attr.title		= NULL;
  html401->textarea.attr.style		= NULL;
}


/*
 * Transitional
 */
/* FONT */
static void
fontOutput(text, size, color)
	const String	text;
	const String	size;
	const String	color;
{
  if (html401->dtd != HTML4_Strict) {
    fputs("<font", stdout);
    if (size != NULL) {
      fprintf(stdout, " size=\"%s\"", size);
    }
    if (color != NULL) {
      fprintf(stdout, " color=\"%s\"", color);
    }
    fprintf(stdout, ">%s</font>", text);
  }
}

/* CENTER */
static void
centerBegin(void)
{
  if (html401->dtd != HTML4_Strict) {
    fputs("<center>\n", stdout);
  }
}
static void
centerEnd(void)
{
  if (html401->dtd != HTML4_Strict) {
    fputs("</center>\n", stdout);
  }
}

/* IFRAME */
static void
iframeBegin(src, width, height)
	const String	src;
	Length		width;
	Length		height;
{
  if (html401->dtd != HTML4_Strict) {
    fputs("<iframe", stdout);
    if (src != NULL) {
      fprintf(stdout, " src=\"%s\"", src);
    }
    if (width != NULL) {
      fprintf(stdout, " width=\"%s\"", width);
    }
    if (height != NULL) {
      fprintf(stdout, " height=\"%s\"", height);
    }
    fputs(" scrolling=", stdout);
    switch (html401->iframe.attr.scrolling) {
      case SCROLLING_YES:
	fputs("\"yes\"", stdout);
	break;
      case SCROLLING_NO:
	fputs("\"no\"", stdout);
	break;
      case SCROLLING_AUTO:
	fputs("\"auto\"", stdout);
      default:
	break;
    }
    fprintf(stdout, " frameborder=\"%d\"",
	    html401->iframe.attr.frameborder ? 1 : 0);
    if (html401->iframe.attr.marginwidth > 0) {
      fprintf(stdout, " marginwidth=\"%d\"", html401->iframe.attr.marginwidth);
    }
    if (html401->iframe.attr.marginheight > 0) {
      fprintf(stdout, " marginheight=\"%d\"",
	      html401->iframe.attr.marginheight);
    }
    printTextAlign(html401->iframe.attr.trans.align);
    fputs(">\n", stdout);
  }
}
static void
iframeEnd(void)
{
  if (html401->dtd != HTML4_Strict) {
    fputs("</iframe>\n", stdout);
  }
}
static void
iframeAttrInit(void)
{
  if (html401->dtd != HTML4_Strict) {
    html401->iframe.attr.scrolling	= SCROLLING_AUTO;
    html401->iframe.attr.frameborder	= TRUE;
    html401->iframe.attr.marginwidth	= 0;
    html401->iframe.attr.marginheight	= 0;
    html401->iframe.attr.trans.align	= ALIGN_NONE;
  }
}


/*
 * Frameset
 */
/* FRAMESET */
static void
framesetBegin(rows, cols)
	const String	rows;
	const String	cols;
{
  if (html401->dtd == HTML4_Frameset) {
    fputs("<frameset", stdout);
    if (rows != NULL) {
      fprintf(stdout, " rows=\"%s\"", rows);
    }
    if (cols != NULL) {
      fprintf(stdout, " cols=\"%s\"", cols);
    }
    fputs(">\n", stdout);
  }
}
static void
framesetEnd(void)
{
  if (html401->dtd == HTML4_Frameset) {
    fputs("</frameset>\n", stdout);
  }
}

/* FRAME */
static void
frameOutput(src)
	const String	src;
{
  if (html401->dtd == HTML4_Frameset) {
    fputs("<iframe", stdout);
    if (src != NULL) {
      fprintf(stdout, " src=\"%s\"", src);
    }
    if (html401->frame.attr.noresize) {
      fputs(" noresize", stdout);
    }
    fputs(" scrolling=", stdout);
    switch (html401->frame.attr.scrolling) {
      case SCROLLING_YES:
	fputs("\"yes\"", stdout);
	break;
      case SCROLLING_NO:
	fputs("\"no\"", stdout);
	break;
      case SCROLLING_AUTO:
	fputs("\"auto\"", stdout);
      default:
	break;
    }
    fprintf(stdout, " frameborder=\"%d\"",
	    html401->frame.attr.frameborder ? 1 : 0);
    if (html401->frame.attr.marginwidth > 0) {
      fprintf(stdout, " marginwidth=\"%d\"", html401->frame.attr.marginwidth);
    }
    if (html401->frame.attr.marginheight > 0) {
      fprintf(stdout, " marginheight=\"%d\"", html401->frame.attr.marginheight);
    }
    fputs(">\n", stdout);
  }
}
static void
frameAttrInit(void)
{
  if (html401->dtd == HTML4_Frameset) {
    html401->frame.attr.noresize	= FALSE;
    html401->frame.attr.scrolling	= SCROLLING_AUTO;
    html401->frame.attr.frameborder	= TRUE;
    html401->frame.attr.marginwidth	= 0;
    html401->frame.attr.marginheight	= 0;
  }
}

/* NOFRAMES */
static void
noframesBegin(void)
{
  if (html401->dtd == HTML4_Frameset) {
    fputs("<noframes>\n", stdout);
  }
}
static void
noframesEnd(void)
{
  if (html401->dtd == HTML4_Frameset) {
    fputs("</noframes>\n", stdout);
  }
}



/*****************************************************************************
 *****************************************************************************/

static void
Print(text)
	const String	text;
{
  if (text != NULL) {
    fputs(text, stdout);
  }
}


/*****************************************************************************
 *****************************************************************************/

static void
printGeneralAttribute(name, value)
	const String	name;
	const String	value;
{
  if (name != NULL && value != NULL) {
    fprintf(stdout, " %s=\"%s\"", name, value);
  }
}

static void
printLinkType(name, linkType)
	const String	name;
	link_t		linkType;
{
  if (name == NULL || linkType == LINK_NONE) {
    return;
  }
  fprintf(stdout, " %s=\"", name);
  switch (linkType) {
    case LINK_ALTERNATE:
      fputs("alternate", stdout);
      break;
    case LINK_STYLESHEET:
      fputs("stylesheet", stdout);
      break;
    case LINK_START:
      fputs("start", stdout);
      break;
    case LINK_NEXT:
      fputs("next", stdout);
      break;
    case LINK_PREV:
      fputs("prev", stdout);
      break;
    case LINK_CONTENTS:
      fputs("contents", stdout);
      break;
    case LINK_INDEX:
      fputs("index", stdout);
      break;
    case LINK_GLOSSARY:
      fputs("glossary", stdout);
      break;
    case LINK_COPYRIGHT:
      fputs("copyright", stdout);
      break;
    case LINK_CHAPTER:
      fputs("chapter", stdout);
      break;
    case LINK_SECTION:
      fputs("section", stdout);
      break;
    case LINK_SUBSECTION:
      fputs("subsection", stdout);
      break;
    case LINK_APPENDIX:
      fputs("appendix", stdout);
      break;
    case LINK_HELP:
      fputs("help", stdout);
      break;
    case LINK_BOOKMARK:
      fputs("bookmark", stdout);
      break;
    default:
      break;
  }
  fputs("\"", stdout);
}

static void
printMediaType(mediaType)
	media_t	mediaType;
{
  if (mediaType == MEDIA_NONE) {
    return;
  }

  fputs(" media=\"", stdout);
  switch (mediaType) {
    case MEDIA_ALL:
      fputs("all", stdout);
      break;
    case MEDIA_AURAL:
      fputs("aural", stdout);
      break;
    case MEDIA_BRAILLE:
      fputs("braille", stdout);
      break;
    case MEDIA_EMBOSSED:
      fputs("embossed", stdout);
      break;
    case MEDIA_HANDHELD:
      fputs("handheld", stdout);
      break;
    case MEDIA_PRINT:
      fputs("print", stdout);
      break;
    case MEDIA_PROJECTION:
      fputs("projection", stdout);
      break;
    case MEDIA_SCREEN:
      fputs("screen", stdout);
      break;
    case MEDIA_TTY:
      fputs("tty", stdout);
      break;
    case MEDIA_TV:
      fputs("tv", stdout);
      break;
    default:
      break;
  }
  fputs("\"", stdout);
}


static void
printTextAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_CENTER:
      fputs(" align=\"center\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    case ALIGN_JUSTIFY:
      fputs(" align=\"justify\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

static void
printHorizontalAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_CENTER:
      fputs(" align=\"center\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

static void
printImageAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_TOP:
      fputs(" align=\"top\"", stdout);
      break;
    case ALIGN_MIDDLE:
      fputs(" align=\"middle\"", stdout);
      break;
    case ALIGN_BOTTOM:
      fputs(" align=\"bottom\"", stdout);
      break;
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

static void
printCaptionAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_TOP:
      fputs(" align=\"top\"", stdout);
      break;
    case ALIGN_BOTTOM:
      fputs(" align=\"bottom\"", stdout);
      break;
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

static void
printTableHorizontalAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_LEFT:
      fputs(" align=\"left\"", stdout);
      break;
    case ALIGN_CENTER:
      fputs(" align=\"center\"", stdout);
      break;
    case ALIGN_RIGHT:
      fputs(" align=\"right\"", stdout);
      break;
    case ALIGN_JUSTIFY:
      fputs(" align=\"justify\"", stdout);
      break;
    case ALIGN_CHAR:
      fputs(" align=\"char\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

static void
printTableVerticalAlign(align)
	align_t	align;
{
  switch (align) {
    case ALIGN_TOP:
      fputs(" align=\"top\"", stdout);
      break;
    case ALIGN_MIDDLE:
      fputs(" align=\"middle\"", stdout);
      break;
    case ALIGN_BOTTOM:
      fputs(" align=\"bottom\"", stdout);
      break;
    case ALIGN_BASELINE:
      fputs(" align=\"baseline\"", stdout);
      break;
    case ALIGN_NONE:
    default:
      break;
  }
}

