/*****************************************************************************
 *
 * FILE:	calendar.c
 * DESCRIPTION:	Calendar CGI
 * DATE:	Sun, Apr  7 2002
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/cover.html
 * COPYRIGHT:	(C) 2002 졿Kouichi ABE (WALL), All rights reserved.
 * COMPILE:	cc calendar.c -o calendar.cgi
 *		-I/usr/local/include -L/usr/local/lib -lcgi -lwkf -lmd
 * $Id: calendar.c,v 1.1 2002/04/06 20:20:17 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <cockatrice.h>


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	LEAP(x)	((((x) % 4 == 0) && ((x) % 100 != 0)) || ((x) % 400 == 0))


/*****************************************************************************
 *
 *	definition of local functions
 *
 *****************************************************************************/
static void	makeCalendar(int, int);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
static CGI *	cgi;	/* cgi object */


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	makeCalendar
 * DESCRIPTION:		make a calendar
 * ARGUMENTS:		year, month
 * RETURN VALUE:	none
 */
static void
makeCalendar(year, month)
	int	year;
	int	month;
{
#define	WDAYS	7
  static String	wday[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
  static int	mday[] = { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  static String	months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

  int		firstDay = cgi->date.getWDay(year, month, 1);
  int		leap	 = LEAP(year);
  String	title;
  int		i;
  bool_t	lieu;
  int		d;

  mday[2] += leap;	/* revise */

  cgi->html->table.attr.border = 0;
  cgi->html->table.begin();
  asprintf(&title, "%s %d", months[month - 1], year);
  cgi->html->CAPTION(title);
  free(title);
  /* print a day of the week */
  cgi->html->tr.begin();
  for (i = 0; i < WDAYS; i++) {
    cgi->html->th.begin();
    cgi->html->printf("%s", wday[i]);
    cgi->html->th.end();
  }
  cgi->html->tr.end();
  /* print a days of a month */
  cgi->html->tr.begin();
  cgi->html->td.attr.align = ALIGN_RIGHT;
  for (i = 0; i < firstDay; i++) {	/* padding */
    cgi->html->td.begin();
    cgi->html->print("&nbsp;");
    cgi->html->td.end();
  }
  for (lieu = FALSE, d = 1; d <= mday[month]; d++, i++) {
    if (i % WDAYS == 0) {
      cgi->html->tr.begin();
    }
    if (cgi->date.isFestival(year, month, d)) {	/* Festival ?*/
      if (cgi->date.getWDay(year, month, d) == 0) {	/* Sunday? */
	lieu = TRUE;
      }
      cgi->html->td.begin();
      cgi->html->printf("$%d", d);
      cgi->html->td.end();
    }
    else {	/* Weekday */
      cgi->html->td.begin();
      if (lieu) {	/* Festival in lieu */
	lieu = FALSE;
	cgi->html->printf("+%d", d);
      }
      else {	/* Weekday */
	switch (i % WDAYS) {
	  case 0:	/* Sunday */
	    cgi->html->printf("*%d", d);
	    break;
	  case 6:	/* Saturday */
	    cgi->html->printf("#%d", d);
	    break;
	  default:
	    cgi->html->printf("%d", d);
	    break;
	}
      }
      cgi->html->td.end();
    }
    if (i % WDAYS == 6) {
      cgi->html->tr.end();
    }
  }
  mday[2] -= leap;	/* end of revised */
  cgi->html->tr.end();
  cgi->html->table.end();
}


int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  int	year;
  int	month;

  cgi = newCGI(HTML4_Transitional);

  if (cgi->param("year") != NULL) {
    year = atoi(cgi->param("year"));
  }
  else {
    year = cgi->date.year();
  }

  if (cgi->param("month") != NULL) {
    month = atoi(cgi->param("month"));
  }
  else {
    month = cgi->date.month();
  }

  cgi->header("text/html");
  cgi->html->html.begin();
    cgi->html->head.begin();
      cgi->html->title("Calendar");
    cgi->html->head.end();
    cgi->html->body.begin();
      cgi->html->H1("Calendar by Cockatrice");
      makeCalendar(year, month);
    cgi->html->body.end();
  cgi->html->html.end();

  cgi->done();

  return 0;
}

